import requests
import gettext
import os
from functools import partial
from contextlib import suppress
from typing import Dict

from requests.structures import CaseInsensitiveDict
from ytmusicapi.auth.headers import load_headers_file, prepare_headers
from ytmusicapi.parsers.i18n import Parser
from ytmusicapi.helpers import *
from ytmusicapi.mixins.browsing import BrowsingMixin
from ytmusicapi.mixins.search import SearchMixin
from ytmusicapi.mixins.watch import WatchMixin
from ytmusicapi.mixins.explore import ExploreMixin
from ytmusicapi.mixins.library import LibraryMixin
from ytmusicapi.mixins.playlists import PlaylistsMixin
from ytmusicapi.mixins.uploads import UploadsMixin
from ytmusicapi.auth.oauth import YTMusicOAuth, is_oauth


class YTMusic(BrowsingMixin, SearchMixin, WatchMixin, ExploreMixin, LibraryMixin, PlaylistsMixin,
              UploadsMixin):
    """
    Allows automated interactions with YouTube Music by emulating the YouTube web client's requests.
    Permits both authenticated and non-authenticated requests.
    Authentication header data must be provided on initialization.
    """

    def __init__(self,
                 auth: str = None,
                 user: str = None,
                 requests_session=True,
                 proxies: dict = None,
                 language: str = 'en',
                 location: str = ''):
        """
        Create a new instance to interact with YouTube Music.

        :param auth: Optional. Provide a string or path to file.
          Authentication credentials are needed to manage your library.
          See :py:func:`setup` for how to fill in the correct credentials.
          Default: A default header is used without authentication.
        :param user: Optional. Specify a user ID string to use in requests.
          This is needed if you want to send requests on behalf of a brand account.
          Otherwise the default account is used. You can retrieve the user ID
          by going to https://myaccount.google.com/brandaccounts and selecting your brand account.
          The user ID will be in the URL: https://myaccount.google.com/b/user_id/
        :param requests_session: A Requests session object or a truthy value to create one.
          Default sessions have a request timeout of 30s, which produces a requests.exceptions.ReadTimeout.
          The timeout can be changed by passing your own Session object::

            s = requests.Session()
            s.request = functools.partial(s.request, timeout=3)
            ytm = YTMusic(session=s)

          A falsy value disables sessions.
          It is generally a good idea to keep sessions enabled for
          performance reasons (connection pooling).
        :param proxies: Optional. Proxy configuration in requests_ format_.

            .. _requests: https://requests.readthedocs.io/
            .. _format: https://requests.readthedocs.io/en/master/user/advanced/#proxies

        :param language: Optional. Can be used to change the language of returned data.
            English will be used by default. Available languages can be checked in
            the ytmusicapi/locales directory.
        :param location: Optional. Can be used to change the location of the user.
            No location will be set by default. This means it is determined by the server.
            Available languages can be checked in the FAQ.
        """
        self.auth = auth
        self.input_dict = None
        self.is_oauth_auth = False

        if isinstance(requests_session, requests.Session):
            self._session = requests_session
        else:
            if requests_session:  # Build a new session.
                self._session = requests.Session()
                self._session.request = partial(self._session.request, timeout=30)
            else:  # Use the Requests API module as a "session".
                self._session = requests.api

        self.proxies = proxies
        # see google cookie docs: https://policies.google.com/technologies/cookies
        # value from https://github.com/yt-dlp/yt-dlp/blob/master/yt_dlp/extractor/youtube.py#L502
        self.cookies = {'SOCS': 'CAI'}
        if self.auth is not None:
            input_json = load_headers_file(self.auth)
            self.input_dict = CaseInsensitiveDict(input_json)
            self.input_dict['filepath'] = self.auth
            self.is_oauth_auth = is_oauth(self.input_dict)

        self.headers = prepare_headers(self._session, proxies, self.input_dict)

        if 'x-goog-visitor-id' not in self.headers:
            self.headers.update(get_visitor_id(self._send_get_request))

        # prepare context
        self.context = initialize_context()

        if location:
            if location not in SUPPORTED_LOCATIONS:
                raise Exception("Location not supported. Check the FAQ for supported locations.")
            self.context['context']['client']['gl'] = location

        if language not in SUPPORTED_LANGUAGES:
            raise Exception("Language not supported. Supported languages are "
                            + (', '.join(SUPPORTED_LANGUAGES)) + ".")
        self.context['context']['client']['hl'] = language
        self.language = language
        try:
            locale.setlocale(locale.LC_ALL, self.language)
        except locale.Error:
            with suppress(locale.Error):
                locale.setlocale(locale.LC_ALL, 'en_US.UTF-8')

        locale_dir = os.path.abspath(os.path.dirname(__file__)) + os.sep + 'locales'
        self.lang = gettext.translation('base', localedir=locale_dir, languages=[language])
        self.parser = Parser(self.lang)

        if user:
            self.context['context']['user']['onBehalfOfUser'] = user

        auth_header = self.headers.get("authorization")
        self.is_browser_auth = auth_header and "SAPISIDHASH" in auth_header
        if self.is_browser_auth:
            try:
                cookie = self.headers.get('cookie')
                self.sapisid = sapisid_from_cookie(cookie)
            except KeyError:
                raise Exception("Your cookie is missing the required value __Secure-3PAPISID")

    def _send_request(self, endpoint: str, body: Dict, additionalParams: str = "") -> Dict:

        if self.is_oauth_auth:
            self.headers = prepare_headers(self._session, self.proxies, self.input_dict)
        body.update(self.context)
        params = YTM_PARAMS
        if self.is_browser_auth:
            origin = self.headers.get('origin', self.headers.get('x-origin'))
            self.headers["authorization"] = get_authorization(self.sapisid + ' ' + origin)
            params += YTM_PARAMS_KEY

        response = self._session.post(YTM_BASE_API + endpoint + params + additionalParams,
                                      json=body,
                                      headers=self.headers,
                                      proxies=self.proxies,
                                      cookies=self.cookies)
        response_text = json.loads(response.text)
        if response.status_code >= 400:
            message = "Server returned HTTP " + str(
                response.status_code) + ": " + response.reason + ".\n"
            error = response_text.get('error', {}).get('message')
            raise Exception(message + error)
        return response_text

    def _send_get_request(self, url: str, params: Dict = None):
        response = self._session.get(url,
                                     params=params,
                                     headers=self.headers,
                                     proxies=self.proxies,
                                     cookies=self.cookies)
        return response

    def _check_auth(self):
        if not self.auth:
            raise Exception("Please provide authentication before using this function")

    def __enter__(self):
        return self

    def __exit__(self, execType=None, execValue=None, trackback=None):
        pass



import argparse
import sys
from pathlib import Path
from typing import Dict

import requests

from ytmusicapi.auth.browser import setup_browser
from ytmusicapi.auth.oauth import YTMusicOAuth


def setup(filepath: str = None, headers_raw: str = None) -> Dict:
    """
    Requests browser headers from the user via command line
    and returns a string that can be passed to YTMusic()

    :param filepath: Optional filepath to store headers to.
    :param headers_raw: Optional request headers copied from browser.
        Otherwise requested from terminal
    :return: configuration headers string
    """
    return setup_browser(filepath, headers_raw)


def setup_oauth(filepath: str = None,
                session: requests.Session = None,
                proxies: dict = None,
                open_browser: bool = False) -> Dict:
    """
    Starts oauth flow from the terminal
    and returns a string that can be passed to YTMusic()

    :param session: Session to use for authentication
    :param proxies: Proxies to use for authentication
    :param filepath: Optional filepath to store headers to.
    :param open_browser: If True, open the default browser with the setup link
    :return: configuration headers string
    """
    if not session:
        session = requests.Session()

    return YTMusicOAuth(session, proxies).setup(filepath, open_browser)


def parse_args(args):
    parser = argparse.ArgumentParser(description='Setup ytmusicapi.')
    parser.add_argument("setup_type",
                        type=str,
                        choices=["oauth", "browser"],
                        help="choose a setup type.")
    parser.add_argument("--file", type=Path, help="optional path to output file.")
    return parser.parse_args(args)


def main():
    args = parse_args(sys.argv[1:])
    filename = args.file.as_posix() if args.file else f"{args.setup_type}.json"
    print(f"Creating {filename} with your authentication credentials...")
    if args.setup_type == "oauth":
        return setup_oauth(filename, open_browser=True)
    else:
        return setup(filename)



import re
import json
from http.cookies import SimpleCookie
from hashlib import sha1
import time
import locale

import unicodedata

from ytmusicapi.constants import *


def initialize_headers():
    return {
        "user-agent": USER_AGENT,
        "accept": "*/*",
        "accept-encoding": "gzip, deflate",
        "content-type": "application/json",
        "content-encoding": "gzip",
        "origin": YTM_DOMAIN
    }


def initialize_context():
    return {
        'context': {
            'client': {
                'clientName': 'WEB_REMIX',
                'clientVersion': '1.' + time.strftime("%Y%m%d", time.gmtime()) + '.01.00'
            },
            'user': {}
        }
    }


def get_visitor_id(request_func):
    response = request_func(YTM_DOMAIN)
    matches = re.findall(r'ytcfg\.set\s*\(\s*({.+?})\s*\)\s*;', response.text)
    visitor_id = ""
    if len(matches) > 0:
        ytcfg = json.loads(matches[0])
        visitor_id = ytcfg.get('VISITOR_DATA')
    return {'X-Goog-Visitor-Id': visitor_id}


def sapisid_from_cookie(raw_cookie):
    cookie = SimpleCookie()
    cookie.load(raw_cookie.replace("\"", ""))
    return cookie['__Secure-3PAPISID'].value


# SAPISID Hash reverse engineered by
# https://stackoverflow.com/a/32065323/5726546
def get_authorization(auth):
    sha_1 = sha1()
    unix_timestamp = str(int(time.time()))
    sha_1.update((unix_timestamp + ' ' + auth).encode('utf-8'))
    return "SAPISIDHASH " + unix_timestamp + "_" + sha_1.hexdigest()


def to_int(string):
    string = unicodedata.normalize("NFKD", string)
    number_string = re.sub(r'\D', '', string)
    try:
        int_value = locale.atoi(number_string)
    except ValueError:
        number_string = number_string.replace(',', '')
        int_value = int(number_string)
    return int_value


def sum_total_duration(item):
    if 'tracks' not in item:
        return 0
    return sum([
        track['duration_seconds'] if 'duration_seconds' in track else 0 for track in item['tracks']
    ])



# commonly used navigation paths
CONTENT = ['contents', 0]
RUN_TEXT = ['runs', 0, 'text']
TAB_CONTENT = ['tabs', 0, 'tabRenderer', 'content']
TAB_1_CONTENT = ['tabs', 1, 'tabRenderer', 'content']
SINGLE_COLUMN = ['contents', 'singleColumnBrowseResultsRenderer']
SINGLE_COLUMN_TAB = SINGLE_COLUMN + TAB_CONTENT
SECTION_LIST = ['sectionListRenderer', 'contents']
SECTION_LIST_ITEM = ['sectionListRenderer'] + CONTENT
ITEM_SECTION = ['itemSectionRenderer'] + CONTENT
MUSIC_SHELF = ['musicShelfRenderer']
GRID = ['gridRenderer']
GRID_ITEMS = GRID + ['items']
MENU = ['menu', 'menuRenderer']
MENU_ITEMS = MENU + ['items']
MENU_LIKE_STATUS = MENU + ['topLevelButtons', 0, 'likeButtonRenderer', 'likeStatus']
MENU_SERVICE = ['menuServiceItemRenderer', 'serviceEndpoint']
TOGGLE_MENU = 'toggleMenuServiceItemRenderer'
PLAY_BUTTON = [
    'overlay', 'musicItemThumbnailOverlayRenderer', 'content', 'musicPlayButtonRenderer'
]
NAVIGATION_BROWSE = ['navigationEndpoint', 'browseEndpoint']
NAVIGATION_BROWSE_ID = NAVIGATION_BROWSE + ['browseId']
PAGE_TYPE = [
    'browseEndpointContextSupportedConfigs', 'browseEndpointContextMusicConfig', 'pageType'
]
WATCH_VIDEO_ID = ['watchEndpoint', 'videoId']
NAVIGATION_VIDEO_ID = ['navigationEndpoint'] + WATCH_VIDEO_ID
QUEUE_VIDEO_ID = ['queueAddEndpoint','queueTarget','videoId']
NAVIGATION_VIDEO_ID_2 = MENU_ITEMS+[0]+MENU_SERVICE+QUEUE_VIDEO_ID
NAVIGATION_PLAYLIST_ID = ['navigationEndpoint', 'watchEndpoint', 'playlistId']
NAVIGATION_WATCH_PLAYLIST_ID = ['navigationEndpoint', 'watchPlaylistEndpoint', 'playlistId']
NAVIGATION_VIDEO_TYPE = [
    'watchEndpoint', 'watchEndpointMusicSupportedConfigs', 'watchEndpointMusicConfig',
    'musicVideoType'
]
TITLE = ['title', 'runs', 0]
TITLE_TEXT = ['title'] + RUN_TEXT
TEXT_RUNS = ['text', 'runs']
TEXT_RUN = TEXT_RUNS + [0]
TEXT_RUN_TEXT = TEXT_RUN + ['text']
SUBTITLE = ['subtitle'] + RUN_TEXT
SUBTITLE_RUNS = ['subtitle', 'runs']
SUBTITLE2 = SUBTITLE_RUNS + [2, 'text']
SUBTITLE3 = SUBTITLE_RUNS + [4, 'text']
THUMBNAIL = ['thumbnail', 'thumbnails']
THUMBNAILS = ['thumbnail', 'musicThumbnailRenderer'] + THUMBNAIL
THUMBNAIL_RENDERER = ['thumbnailRenderer', 'musicThumbnailRenderer'] + THUMBNAIL
THUMBNAIL_CROPPED = ['thumbnail', 'croppedSquareThumbnailRenderer'] + THUMBNAIL
FEEDBACK_TOKEN = ['feedbackEndpoint', 'feedbackToken']
BADGE_PATH = [0, 'musicInlineBadgeRenderer', 'accessibilityData', 'accessibilityData', 'label']
BADGE_LABEL = ['badges'] + BADGE_PATH
SUBTITLE_BADGE_LABEL = ['subtitleBadges'] + BADGE_PATH
CATEGORY_TITLE = ['musicNavigationButtonRenderer', 'buttonText'] + RUN_TEXT
CATEGORY_PARAMS = ['musicNavigationButtonRenderer', 'clickCommand', 'browseEndpoint', 'params']
MRLIR = 'musicResponsiveListItemRenderer'
MTRIR = 'musicTwoRowItemRenderer'
TASTE_PROFILE_ITEMS = ["contents", "tastebuilderRenderer", "contents"]
TASTE_PROFILE_ARTIST = ["title", "runs"]
SECTION_LIST_CONTINUATION = ['continuationContents', 'sectionListContinuation']
MENU_PLAYLIST_ID = MENU_ITEMS + [0, 'menuNavigationItemRenderer'] + NAVIGATION_WATCH_PLAYLIST_ID
HEADER_DETAIL = ['header', 'musicDetailHeaderRenderer']
DESCRIPTION_SHELF = ['musicDescriptionShelfRenderer']
DESCRIPTION = ['description'] + RUN_TEXT
CAROUSEL = ['musicCarouselShelfRenderer']
IMMERSIVE_CAROUSEL = ['musicImmersiveCarouselShelfRenderer']
CAROUSEL_CONTENTS = CAROUSEL + ['contents']
CAROUSEL_TITLE = ['header', 'musicCarouselShelfBasicHeaderRenderer'] + TITLE
CARD_SHELF_TITLE = ['header', 'musicCardShelfHeaderBasicRenderer'] + TITLE_TEXT
FRAMEWORK_MUTATIONS = ['frameworkUpdates', 'entityBatchUpdate', 'mutations']


def nav(root, items, none_if_absent=False):
    """Access a nested object in root by item sequence."""
    try:
        for k in items:
            root = root[k]
        return root
    except Exception as err:
        if none_if_absent:
            return None
        else:
            raise err


def find_object_by_key(object_list, key, nested=None, is_key=False):
    for item in object_list:
        if nested:
            item = item[nested]
        if key in item:
            return item[key] if is_key else item
    return None


def find_objects_by_key(object_list, key, nested=None):
    objects = []
    for item in object_list:
        if nested:
            item = item[nested]
        if key in item:
            objects.append(item)
    return objects



YTM_DOMAIN = 'https://music.youtube.com'
YTM_BASE_API = YTM_DOMAIN + '/youtubei/v1/'
YTM_PARAMS = '?alt=json'
YTM_PARAMS_KEY = "&key=AIzaSyC9XL3ZjWddXya6X74dJoCTL-WEYFDNX30"
USER_AGENT = 'Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:88.0) Gecko/20100101 Firefox/88.0'
SUPPORTED_LANGUAGES = {
    'ar', 'de', 'en', 'es', 'fr', 'hi', 'it', 'ja', 'ko', 'nl', 'pt', 'ru', 'tr', 'ur', 'zh_CN',
    'zh_TW'
}
SUPPORTED_LOCATIONS = {
    'AE', 'AR', 'AT', 'AU', 'AZ', 'BA', 'BD', 'BE', 'BG', 'BH', 'BO', 'BR', 'BY', 'CA', 'CH', 'CL',
    'CO', 'CR', 'CY', 'CZ', 'DE', 'DK', 'DO', 'DZ', 'EC', 'EE', 'EG', 'ES', 'FI', 'FR', 'GB', 'GE',
    'GH', 'GR', 'GT', 'HK', 'HN', 'HR', 'HU', 'ID', 'IE', 'IL', 'IN', 'IQ', 'IS', 'IT', 'JM', 'JO',
    'JP', 'KE', 'KH', 'KR', 'KW', 'KZ', 'LA', 'LB', 'LI', 'LK', 'LT', 'LU', 'LV', 'LY', 'MA', 'ME',
    'MK', 'MT', 'MX', 'MY', 'NG', 'NI', 'NL', 'NO', 'NP', 'NZ', 'OM', 'PA', 'PE', 'PG', 'PH', 'PK',
    'PL', 'PR', 'PT', 'PY', 'QA', 'RO', 'RS', 'RU', 'SA', 'SE', 'SG', 'SI', 'SK', 'SN', 'SV', 'TH',
    'TN', 'TR', 'TW', 'TZ', 'UA', 'UG', 'US', 'UY', 'VE', 'VN', 'YE', 'ZA', 'ZW'
}
OAUTH_CLIENT_ID = "861556708454-d6dlm3lh05idd8npek18k6be8ba3oc68.apps.googleusercontent.com"
OAUTH_CLIENT_SECRET = "SboVhoG9s0rNafixCSGGKXAT"
OAUTH_SCOPE = "https://www.googleapis.com/auth/youtube"
OAUTH_CODE_URL = "https://www.youtube.com/o/oauth2/device/code"
OAUTH_TOKEN_URL = "https://oauth2.googleapis.com/token"
OAUTH_USER_AGENT = USER_AGENT + " Cobalt/Version"



import configparser
import json
import time
import unittest
import warnings
from pathlib import Path
from unittest import mock

from requests import Response

from ytmusicapi.setup import main, setup  # noqa: E402
from ytmusicapi.ytmusic import YTMusic  # noqa: E402


def get_resource(file: str) -> str:
    data_dir = Path(__file__).parent
    return data_dir.joinpath(file).as_posix()


config = configparser.RawConfigParser()
config.read(get_resource("test.cfg"), "utf-8")

sample_album = "MPREb_4pL8gzRtw1p"  # Eminem - Revival
sample_video = "hpSrLjc5SMs"  # Oasis - Wonderwall
sample_playlist = "PL6bPxvf5dW5clc3y9wAoslzqUrmkZ5c-u"  # very large playlist

headers_oauth = get_resource(config["auth"]["headers_oauth"])
headers_browser = get_resource(config["auth"]["headers_file"])


class TestYTMusic(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        warnings.filterwarnings(action="ignore", message="unclosed", category=ResourceWarning)
        with YTMusic(requests_session=False) as yt:
            assert isinstance(yt, YTMusic)
        cls.yt = YTMusic()
        cls.yt_oauth = YTMusic(headers_oauth)
        cls.yt_auth = YTMusic(headers_browser, location="GB")
        cls.yt_brand = YTMusic(config["auth"]["headers"], config["auth"]["brand_account"])
        cls.yt_empty = YTMusic(config["auth"]["headers_empty"],
                               config["auth"]["brand_account_empty"])

    @mock.patch("sys.argv", ["ytmusicapi", "browser", "--file", headers_browser])
    def test_setup_browser(self):
        headers = setup(headers_browser, config["auth"]["headers_raw"])
        self.assertGreaterEqual(len(headers), 2)
        headers_raw = config["auth"]["headers_raw"].split("\n")
        with mock.patch("builtins.input", side_effect=(headers_raw + [EOFError()])):
            headers = main()
            self.assertGreaterEqual(len(headers), 2)

    @mock.patch("requests.Response.json")
    @mock.patch("requests.Session.post")
    @mock.patch("sys.argv", ["ytmusicapi", "oauth", "--file", headers_oauth])
    def test_setup_oauth(self, session_mock, json_mock):
        session_mock.return_value = Response()
        json_mock.side_effect = [
            json.loads(config["auth"]["oauth_code"]),
            json.loads(config["auth"]["oauth_token"]),
        ]
        with mock.patch("builtins.input", return_value="y"):
            main()
            self.assertTrue(Path(headers_oauth).exists())

        json_mock.side_effect = None
        with open(headers_oauth, mode="r", encoding="utf8") as headers:
            string_headers = headers.read()
            self.yt_oauth = YTMusic(string_headers)

    ###############
    # BROWSING
    ###############

    def test_get_home(self):
        result = self.yt.get_home(limit=6)
        self.assertGreaterEqual(len(result), 6)
        result = self.yt_auth.get_home(limit=15)
        self.assertGreaterEqual(len(result), 15)

    def test_search(self):
        query = "edm playlist"
        self.assertRaises(Exception, self.yt_auth.search, query, filter="song")
        self.assertRaises(Exception, self.yt_auth.search, query, scope="upload")
        queries = ["Monekes", "qllwlwl", "heun"]
        for q in queries:
            with self.subTest():
                results = self.yt_brand.search(q)
                self.assertListEqual(["resultType" in r for r in results], [True] * len(results))
                self.assertGreaterEqual(len(results), 10)
                results = self.yt.search(q)
                self.assertGreaterEqual(len(results), 10)
        results = self.yt_auth.search("Martin Stig Andersen - Deteriation", ignore_spelling=True)
        self.assertGreater(len(results), 0)

    def test_search_filters(self):
        query = "hip hop playlist"
        results = self.yt_auth.search(query, filter="songs")
        self.assertGreater(len(results), 10)
        self.assertTrue(all(item['resultType'] == 'song' for item in results))
        results = self.yt_auth.search(query, filter="videos")
        self.assertGreater(len(results), 10)
        self.assertTrue(all(item['resultType'] == 'video' for item in results))
        results = self.yt_auth.search(query, filter="albums", limit=40)
        self.assertGreater(len(results), 20)
        self.assertTrue(all(item['resultType'] == 'album' for item in results))
        results = self.yt_auth.search("project-2", filter="artists", ignore_spelling=True)
        self.assertGreater(len(results), 10)
        self.assertTrue(all(item['resultType'] == 'artist' for item in results))
        results = self.yt_auth.search("classical music", filter="playlists")
        self.assertGreater(len(results), 10)
        self.assertTrue(all(item['resultType'] == 'playlist' for item in results))
        results = self.yt_auth.search("clasical music", filter="playlists", ignore_spelling=True)
        self.assertGreater(len(results), 10)
        results = self.yt_auth.search("clasic rock",
                                      filter="community_playlists",
                                      ignore_spelling=True)
        self.assertGreater(len(results), 10)
        self.assertTrue(all(item['resultType'] == 'playlist' for item in results))
        results = self.yt_auth.search("hip hop", filter="featured_playlists")
        self.assertGreater(len(results), 10)
        self.assertTrue(all(item['resultType'] == 'playlist' for item in results))
        results = self.yt_auth.search("some user", filter="profiles")
        self.assertGreater(len(results), 10)
        self.assertTrue(all(item['resultType'] == 'profile' for item in results))

    def test_search_uploads(self):
        self.assertRaises(
            Exception,
            self.yt.search,
            "audiomachine",
            filter="songs",
            scope="uploads",
            limit=40,
        )
        results = self.yt_auth.search("audiomachine", scope="uploads", limit=40)
        self.assertGreater(len(results), 20)

    def test_search_library(self):
        results = self.yt_oauth.search("garrix", scope="library")
        self.assertGreater(len(results), 5)
        results = self.yt_auth.search("bergersen", filter="songs", scope="library", limit=40)
        self.assertGreater(len(results), 10)
        results = self.yt_auth.search("garrix", filter="albums", scope="library", limit=40)
        self.assertGreaterEqual(len(results), 4)
        results = self.yt_auth.search("garrix", filter="artists", scope="library", limit=40)
        self.assertGreaterEqual(len(results), 1)
        results = self.yt_auth.search("garrix", filter="playlists", scope="library")
        self.assertGreaterEqual(len(results), 1)
        self.assertRaises(Exception,
                          self.yt_auth.search,
                          "beatles",
                          filter="community_playlists",
                          scope="library",
                          limit=40)
        self.assertRaises(Exception,
                          self.yt_auth.search,
                          "beatles",
                          filter="featured_playlists",
                          scope="library",
                          limit=40)

    def test_get_artist(self):
        results = self.yt.get_artist("MPLAUCmMUZbaYdNH0bEd1PAlAqsA")
        self.assertEqual(len(results), 14)

        # test correctness of related artists
        related = results["related"]["results"]
        self.assertEqual(
            len([
                x for x in related
                if set(x.keys()) == {"browseId", "subscribers", "title", "thumbnails"}
            ]),
            len(related),
        )

        results = self.yt.get_artist("UCLZ7tlKC06ResyDmEStSrOw")  # no album year
        self.assertGreaterEqual(len(results), 11)

    def test_get_artist_albums(self):
        artist = self.yt.get_artist("UCj5ZiBBqpe0Tg4zfKGHEFuQ")
        results = self.yt.get_artist_albums(artist["albums"]["browseId"],
                                            artist["albums"]["params"])
        self.assertGreater(len(results), 0)

    def test_get_artist_singles(self):
        artist = self.yt.get_artist("UCAeLFBCQS7FvI8PvBrWvSBg")
        results = self.yt.get_artist_albums(artist["singles"]["browseId"],
                                            artist["singles"]["params"])
        self.assertGreater(len(results), 0)

    def test_get_user(self):
        results = self.yt.get_user("UC44hbeRoCZVVMVg5z0FfIww")
        self.assertEqual(len(results), 3)

    def test_get_user_playlists(self):
        results = self.yt.get_user("UCPVhZsC2od1xjGhgEc2NEPQ")
        results = self.yt.get_user_playlists("UCPVhZsC2od1xjGhgEc2NEPQ",
                                             results["playlists"]["params"])
        self.assertGreater(len(results), 100)

    def test_get_album_browse_id(self):
        browse_id = self.yt.get_album_browse_id("OLAK5uy_nMr9h2VlS-2PULNz3M3XVXQj_P3C2bqaY")
        self.assertEqual(browse_id, sample_album)

    def test_get_album(self):
        results = self.yt_auth.get_album(sample_album)
        self.assertGreaterEqual(len(results), 9)
        self.assertTrue(results["tracks"][0]["isExplicit"])
        self.assertIn("feedbackTokens", results["tracks"][0])
        self.assertEqual(len(results["other_versions"]), 2)
        results = self.yt.get_album("MPREb_BQZvl3BFGay")
        self.assertEqual(len(results["tracks"]), 7)
        self.assertEqual(len(results["tracks"][0]["artists"]), 1)
        results = self.yt.get_album("MPREb_rqH94Zr3NN0")
        self.assertEqual(len(results["tracks"][0]["artists"]), 2)

    def test_get_song(self):
        song = self.yt_oauth.get_song(config["uploads"]["private_upload_id"])  # private upload
        self.assertEqual(len(song), 5)
        song = self.yt.get_song(sample_video)
        self.assertGreaterEqual(len(song["streamingData"]["adaptiveFormats"]), 10)

    def test_get_song_related_content(self):
        song = self.yt_oauth.get_watch_playlist(sample_video)
        song = self.yt_oauth.get_song_related(song["related"])
        self.assertGreaterEqual(len(song), 5)

    def test_get_lyrics(self):
        playlist = self.yt.get_watch_playlist(sample_video)
        lyrics_song = self.yt.get_lyrics(playlist["lyrics"])
        self.assertIsNotNone(lyrics_song["lyrics"])
        self.assertIsNotNone(lyrics_song["source"])

        playlist = self.yt.get_watch_playlist(config["uploads"]["private_upload_id"])
        self.assertIsNone(playlist["lyrics"])
        self.assertRaises(Exception, self.yt.get_lyrics, playlist["lyrics"])

    def test_get_signatureTimestamp(self):
        signatureTimestamp = self.yt.get_signatureTimestamp()
        self.assertIsNotNone(signatureTimestamp)

    def test_set_tasteprofile(self):
        self.assertRaises(Exception, self.yt.set_tasteprofile, "not an artist")
        taste_profile = self.yt.get_tasteprofile()
        self.assertIsNone(self.yt.set_tasteprofile(list(taste_profile)[:5], taste_profile))

        self.assertRaises(Exception, self.yt_brand.set_tasteprofile, ["test", "test2"])
        taste_profile = self.yt_brand.get_tasteprofile()
        self.assertIsNone(self.yt_brand.set_tasteprofile(list(taste_profile)[:1], taste_profile))

    def test_get_tasteprofile(self):
        result = self.yt.get_tasteprofile()
        self.assertGreaterEqual(len(result), 0)

        result = self.yt_oauth.get_tasteprofile()
        self.assertGreaterEqual(len(result), 0)

    def test_get_search_suggestions(self):
        result = self.yt.get_search_suggestions("fade")
        self.assertGreaterEqual(len(result), 0)

        result = self.yt.get_search_suggestions("fade", detailed_runs=True)
        self.assertGreaterEqual(len(result), 0)

    ################
    # EXPLORE
    ################

    def test_get_mood_playlists(self):
        categories = self.yt.get_mood_categories()
        self.assertGreater(len(list(categories)), 0)
        cat = list(categories)[0]
        self.assertGreater(len(categories[cat]), 0)
        playlists = self.yt.get_mood_playlists(categories[cat][0]["params"])
        self.assertGreater(len(playlists), 0)

    def test_get_charts(self):
        charts = self.yt_oauth.get_charts()
        self.assertEqual(len(charts), 4)
        charts = self.yt.get_charts(country="US")
        self.assertEqual(len(charts), 5)
        charts = self.yt.get_charts(country="BE")
        self.assertEqual(len(charts), 4)

    ###############
    # WATCH
    ###############

    def test_get_watch_playlist(self):
        playlist = self.yt_oauth.get_watch_playlist(
            playlistId="RDAMPLOLAK5uy_l_fKDQGOUsk8kbWsm9s86n4-nZNd2JR8Q",
            radio=True,
            limit=90,
        )
        self.assertGreaterEqual(len(playlist["tracks"]), 90)
        playlist = self.yt_oauth.get_watch_playlist("9mWr4c_ig54", limit=50)
        self.assertGreater(len(playlist["tracks"]), 45)
        playlist = self.yt_oauth.get_watch_playlist("UoAf_y9Ok4k")  # private track
        self.assertGreaterEqual(len(playlist["tracks"]), 25)
        playlist = self.yt.get_watch_playlist(
            playlistId="OLAK5uy_lKgoGvlrWhX0EIPavQUXxyPed8Cj38AWc", shuffle=True)
        self.assertEqual(len(playlist["tracks"]), 12)
        playlist = self.yt_brand.get_watch_playlist(playlistId=config["playlists"]["own"],
                                                    shuffle=True)
        self.assertEqual(len(playlist["tracks"]), 4)

    ################
    # LIBRARY
    ################

    def test_get_library_playlists(self):
        playlists = self.yt_oauth.get_library_playlists(50)
        self.assertGreater(len(playlists), 25)

        playlists = self.yt_auth.get_library_playlists(None)
        self.assertGreaterEqual(len(playlists), config.getint("limits", "library_playlists"))

        playlists = self.yt_empty.get_library_playlists()
        self.assertLessEqual(len(playlists), 1)  # "Episodes saved for later"

    def test_get_library_songs(self):
        self.assertRaises(Exception, self.yt_auth.get_library_songs, None, True)
        songs = self.yt_oauth.get_library_songs(100)
        self.assertGreaterEqual(len(songs), 100)
        songs = self.yt_auth.get_library_songs(200, validate_responses=True)
        self.assertGreaterEqual(len(songs), config.getint("limits", "library_songs"))
        songs = self.yt_auth.get_library_songs(order="a_to_z")
        self.assertGreaterEqual(len(songs), 25)
        songs = self.yt_empty.get_library_songs()
        self.assertEqual(len(songs), 0)

    def test_get_library_albums(self):
        albums = self.yt_oauth.get_library_albums(100)
        self.assertGreater(len(albums), 50)
        for album in albums:
            self.assertIn("playlistId", album)
        albums = self.yt_brand.get_library_albums(100, order="a_to_z")
        self.assertGreater(len(albums), 50)
        albums = self.yt_brand.get_library_albums(100, order="z_to_a")
        self.assertGreater(len(albums), 50)
        albums = self.yt_brand.get_library_albums(100, order="recently_added")
        self.assertGreater(len(albums), 50)
        albums = self.yt_empty.get_library_albums()
        self.assertEqual(len(albums), 0)

    def test_get_library_artists(self):
        artists = self.yt_auth.get_library_artists(50)
        self.assertGreater(len(artists), 40)
        artists = self.yt_oauth.get_library_artists(order="a_to_z", limit=50)
        self.assertGreater(len(artists), 40)
        artists = self.yt_brand.get_library_artists(limit=None)
        self.assertGreater(len(artists), config.getint("limits", "library_artists"))
        artists = self.yt_empty.get_library_artists()
        self.assertEqual(len(artists), 0)

    def test_get_library_subscriptions(self):
        artists = self.yt_brand.get_library_subscriptions(50)
        self.assertGreater(len(artists), 40)
        artists = self.yt_brand.get_library_subscriptions(order="z_to_a")
        self.assertGreater(len(artists), 20)
        artists = self.yt_brand.get_library_subscriptions(limit=None)
        self.assertGreater(len(artists), config.getint("limits", "library_subscriptions"))
        artists = self.yt_empty.get_library_subscriptions()
        self.assertEqual(len(artists), 0)

    def test_get_liked_songs(self):
        songs = self.yt_brand.get_liked_songs(200)
        self.assertGreater(len(songs["tracks"]), 100)
        songs = self.yt_empty.get_liked_songs()
        self.assertEqual(songs["trackCount"], 0)

    def test_get_history(self):
        songs = self.yt_oauth.get_history()
        self.assertGreater(len(songs), 0)

    def test_manipulate_history_items(self):
        song = self.yt_auth.get_song(sample_video)
        response = self.yt_auth.add_history_item(song)
        self.assertEqual(response.status_code, 204)
        songs = self.yt_auth.get_history()
        self.assertGreater(len(songs), 0)
        response = self.yt_auth.remove_history_items([songs[0]["feedbackToken"]])
        self.assertIn("feedbackResponses", response)

    def test_rate_song(self):
        response = self.yt_auth.rate_song(sample_video, "LIKE")
        self.assertIn("actions", response)
        response = self.yt_auth.rate_song(sample_video, "INDIFFERENT")
        self.assertIn("actions", response)

    def test_edit_song_library_status(self):
        album = self.yt_brand.get_album(sample_album)
        response = self.yt_brand.edit_song_library_status(
            album["tracks"][2]["feedbackTokens"]["add"])
        self.assertTrue(response["feedbackResponses"][0]["isProcessed"])
        response = self.yt_brand.edit_song_library_status(
            album["tracks"][2]["feedbackTokens"]["remove"])
        self.assertTrue(response["feedbackResponses"][0]["isProcessed"])

    def test_rate_playlist(self):
        response = self.yt_auth.rate_playlist("OLAK5uy_l3g4WcHZsEx_QuEDZzWEiyFzZl6pL0xZ4", "LIKE")
        self.assertIn("actions", response)
        response = self.yt_auth.rate_playlist("OLAK5uy_l3g4WcHZsEx_QuEDZzWEiyFzZl6pL0xZ4",
                                              "INDIFFERENT")
        self.assertIn("actions", response)

    def test_subscribe_artists(self):
        self.yt_auth.subscribe_artists(["UCUDVBtnOQi4c7E8jebpjc9Q", "UCiMhD4jzUqG-IgPzUmmytRQ"])
        self.yt_auth.unsubscribe_artists(["UCUDVBtnOQi4c7E8jebpjc9Q", "UCiMhD4jzUqG-IgPzUmmytRQ"])

    ###############
    # PLAYLISTS
    ###############

    def test_get_playlist_foreign(self):
        self.assertRaises(Exception, self.yt.get_playlist, "PLABC")
        playlist = self.yt.get_playlist("PLk5BdzXBUiUe8Q5I13ZSCD8HbxMqJUUQA",
                                        limit=300,
                                        suggestions_limit=7)
        self.assertGreater(len(playlist['duration']), 5)
        self.assertGreater(len(playlist["tracks"]), 200)
        self.assertNotIn("suggestions", playlist)

        self.yt.get_playlist("RDATgXd-")
        self.assertGreaterEqual(len(playlist["tracks"]), 100)

        playlist = self.yt_oauth.get_playlist("PLj4BSJLnVpNyIjbCWXWNAmybc97FXLlTk",
                                              limit=None,
                                              related=True)
        self.assertGreater(len(playlist["tracks"]), 200)
        self.assertEqual(len(playlist["related"]), 0)

    def test_get_playlist_owned(self):
        playlist = self.yt_brand.get_playlist(config["playlists"]["own"],
                                              related=True,
                                              suggestions_limit=21)
        self.assertLess(len(playlist["tracks"]), 100)
        self.assertEqual(len(playlist["suggestions"]), 21)
        self.assertEqual(len(playlist["related"]), 10)

    def test_edit_playlist(self):
        playlist = self.yt_brand.get_playlist(config["playlists"]["own"])
        response = self.yt_brand.edit_playlist(
            playlist["id"],
            title="",
            description="",
            privacyStatus="PRIVATE",
            moveItem=(
                playlist["tracks"][1]["setVideoId"],
                playlist["tracks"][0]["setVideoId"],
            ),
        )
        self.assertEqual(response, "STATUS_SUCCEEDED", "Playlist edit failed")
        self.yt_brand.edit_playlist(
            playlist["id"],
            title=playlist["title"],
            description=playlist["description"],
            privacyStatus=playlist["privacy"],
            moveItem=(
                playlist["tracks"][0]["setVideoId"],
                playlist["tracks"][1]["setVideoId"],
            ),
        )
        self.assertEqual(response, "STATUS_SUCCEEDED", "Playlist edit failed")

    # end to end test adding playlist, adding item, deleting item, deleting playlist
    def test_end2end(self):
        playlistId = self.yt_brand.create_playlist(
            "test",
            "test description",
            source_playlist="OLAK5uy_lGQfnMNGvYCRdDq9ZLzJV2BJL2aHQsz9Y",
        )
        self.assertEqual(len(playlistId), 34, "Playlist creation failed")
        self.yt_brand.edit_playlist(playlistId, addToTop=True)
        response = self.yt_brand.add_playlist_items(
            playlistId,
            [sample_video, sample_video],
            source_playlist="OLAK5uy_nvjTE32aFYdFN7HCyMv3cGqD3wqBb4Jow",
            duplicates=True,
        )
        self.assertEqual(response["status"], "STATUS_SUCCEEDED", "Adding playlist item failed")
        self.assertGreater(len(response["playlistEditResults"]), 0, "Adding playlist item failed")
        time.sleep(2)
        self.yt_brand.edit_playlist(playlistId, addToTop=False)
        playlist = self.yt_brand.get_playlist(playlistId, related=True)
        self.assertEqual(len(playlist["tracks"]), 46, "Getting playlist items failed")
        response = self.yt_brand.remove_playlist_items(playlistId, playlist["tracks"])
        self.assertEqual(response, "STATUS_SUCCEEDED", "Playlist item removal failed")
        self.yt_brand.delete_playlist(playlistId)

    ###############
    # UPLOADS
    ###############

    def test_get_library_upload_songs(self):
        results = self.yt_oauth.get_library_upload_songs(50, order="z_to_a")
        self.assertGreater(len(results), 25)

        results = self.yt_empty.get_library_upload_songs(100)
        self.assertEqual(len(results), 0)

    def test_get_library_upload_albums(self):
        results = self.yt_oauth.get_library_upload_albums(50, order="a_to_z")
        self.assertGreater(len(results), 40)

        albums = self.yt_auth.get_library_upload_albums(None)
        self.assertGreaterEqual(len(albums), config.getint("limits", "library_upload_albums"))

        results = self.yt_empty.get_library_upload_albums(100)
        self.assertEqual(len(results), 0)

    def test_get_library_upload_artists(self):
        artists = self.yt_oauth.get_library_upload_artists(None)
        self.assertGreaterEqual(len(artists), config.getint("limits", "library_upload_artists"))

        results = self.yt_auth.get_library_upload_artists(50, order="recently_added")
        self.assertGreaterEqual(len(results), 25)

        results = self.yt_empty.get_library_upload_artists(100)
        self.assertEqual(len(results), 0)

    def test_upload_song(self):
        self.assertRaises(Exception, self.yt_auth.upload_song, "song.wav")
        self.assertRaises(Exception, self.yt_oauth.upload_song, config["uploads"]["file"])
        response = self.yt_auth.upload_song(get_resource(config["uploads"]["file"]))
        self.assertEqual(response.status_code, 409)

    @unittest.skip("Do not delete uploads")
    def test_delete_upload_entity(self):
        results = self.yt_oauth.get_library_upload_songs()
        response = self.yt_oauth.delete_upload_entity(results[0]["entityId"])
        self.assertEqual(response, "STATUS_SUCCEEDED")

    def test_get_library_upload_album(self):
        album = self.yt_oauth.get_library_upload_album(config["uploads"]["private_album_id"])
        self.assertGreater(len(album["tracks"]), 0)

    def test_get_library_upload_artist(self):
        tracks = self.yt_oauth.get_library_upload_artist(config["uploads"]["private_artist_id"],
                                                         100)
        self.assertGreater(len(tracks), 0)


if __name__ == "__main__":
    unittest.main()



from .songs import *
from ._utils import *


def get_search_result_type(result_type_local, result_types_local):
    if not result_type_local:
        return None
    result_types = ['artist', 'playlist', 'song', 'video', 'station', 'profile']
    result_type_local = result_type_local.lower()
    # default to album since it's labeled with multiple values ('Single', 'EP', etc.)
    if result_type_local not in result_types_local:
        result_type = 'album'
    else:
        result_type = result_types[result_types_local.index(result_type_local)]

    return result_type


def parse_top_result(data, search_result_types):
    result_type = get_search_result_type(nav(data, SUBTITLE), search_result_types)
    search_result = {'category': nav(data, CARD_SHELF_TITLE), 'resultType': result_type}
    if result_type == 'artist':
        subscribers = nav(data, SUBTITLE2, True)
        if subscribers:
            search_result['subscribers'] = subscribers.split(' ')[0]

        artist_info = parse_song_runs(nav(data, ['title', 'runs']))
        search_result.update(artist_info)

    if result_type in ['song', 'video']:
        on_tap = data.get('onTap')
        if on_tap:
            search_result['videoId'] = nav(on_tap, WATCH_VIDEO_ID)
            search_result['videoType'] = nav(on_tap, NAVIGATION_VIDEO_TYPE)

    if result_type in ['song', 'video', 'album']:
        search_result['title'] = nav(data, TITLE_TEXT)
        runs = nav(data, ['subtitle', 'runs'])[2:]
        song_info = parse_song_runs(runs)
        search_result.update(song_info)

    search_result['thumbnails'] = nav(data, THUMBNAILS, True)
    return search_result


def parse_search_result(data, search_result_types, result_type, category):
    default_offset = (not result_type or result_type == "album") * 2
    search_result = {'category': category}
    video_type = nav(data, PLAY_BUTTON + ['playNavigationEndpoint'] + NAVIGATION_VIDEO_TYPE, True)
    if not result_type and video_type:
        result_type = 'song' if video_type == 'MUSIC_VIDEO_TYPE_ATV' else 'video'

    result_type = get_search_result_type(get_item_text(data, 1),
                                         search_result_types) if not result_type else result_type
    search_result['resultType'] = result_type

    if result_type != 'artist':
        search_result['title'] = get_item_text(data, 0)

    if result_type == 'artist':
        search_result['artist'] = get_item_text(data, 0)
        parse_menu_playlists(data, search_result)

    elif result_type == 'album':
        search_result['type'] = get_item_text(data, 1)

    elif result_type == 'playlist':
        flex_item = get_flex_column_item(data, 1)['text']['runs']
        has_author = len(flex_item) == default_offset + 3
        search_result['itemCount'] = get_item_text(data, 1,
                                                   default_offset + has_author * 2).split(' ')[0]
        search_result['author'] = None if not has_author else get_item_text(
            data, 1, default_offset)

    elif result_type == 'station':
        search_result['videoId'] = nav(data, NAVIGATION_VIDEO_ID)
        search_result['playlistId'] = nav(data, NAVIGATION_PLAYLIST_ID)

    elif result_type == 'profile':
        search_result['name'] = get_item_text(data, 1, 2, True)

    elif result_type == 'song':
        search_result['album'] = None
        if 'menu' in data:
            toggle_menu = find_object_by_key(nav(data, MENU_ITEMS), TOGGLE_MENU)
            if toggle_menu:
                search_result['feedbackTokens'] = parse_song_menu_tokens(toggle_menu)

    elif result_type == 'upload':
        browse_id = nav(data, NAVIGATION_BROWSE_ID, True)
        if not browse_id:  # song result
            flex_items = [
                nav(get_flex_column_item(data, i), ['text', 'runs'], True) for i in range(2)
            ]
            if flex_items[0]:
                search_result['videoId'] = nav(flex_items[0][0], NAVIGATION_VIDEO_ID, True)
                search_result['playlistId'] = nav(flex_items[0][0], NAVIGATION_PLAYLIST_ID, True)
            if flex_items[1]:
                search_result.update(parse_song_runs(flex_items[1]))
            search_result['resultType'] = 'song'

        else:  # artist or album result
            search_result['browseId'] = browse_id
            if 'artist' in search_result['browseId']:
                search_result['resultType'] = 'artist'
            else:
                flex_item2 = get_flex_column_item(data, 1)
                runs = [
                    run['text'] for i, run in enumerate(flex_item2['text']['runs']) if i % 2 == 0
                ]
                if len(runs) > 1:
                    search_result['artist'] = runs[1]
                if len(runs) > 2:  # date may be missing
                    search_result['releaseDate'] = runs[2]
                search_result['resultType'] = 'album'

    if result_type in ['song', 'video']:
        search_result['videoId'] = nav(
            data, PLAY_BUTTON + ['playNavigationEndpoint', 'watchEndpoint', 'videoId'], True)
        search_result['videoType'] = video_type

    if result_type in ['song', 'video', 'album']:
        search_result['duration'] = None
        search_result['year'] = None
        flex_item = get_flex_column_item(data, 1)
        runs = flex_item['text']['runs'][default_offset:]
        song_info = parse_song_runs(runs)
        search_result.update(song_info)

    if result_type in ['artist', 'album', 'playlist', 'profile']:
        search_result['browseId'] = nav(data, NAVIGATION_BROWSE_ID, True)

    if result_type in ['song', 'album']:
        search_result['isExplicit'] = nav(data, BADGE_LABEL, True) is not None

    search_result['thumbnails'] = nav(data, THUMBNAILS, True)

    return search_result


def parse_search_results(results, search_result_types, resultType=None, category=None):
    return [
        parse_search_result(result[MRLIR], search_result_types, resultType, category)
        for result in results
    ]


def get_search_params(filter, scope, ignore_spelling):
    filtered_param1 = 'EgWKAQ'
    params = None
    if filter is None and scope is None and not ignore_spelling:
        return params

    if scope == 'uploads':
        params = 'agIYAw%3D%3D'

    if scope == 'library':
        if filter:
            param1 = filtered_param1
            param2 = _get_param2(filter)
            param3 = 'AWoKEAUQCRADEAoYBA%3D%3D'
        else:
            params = 'agIYBA%3D%3D'

    if scope is None and filter:
        if filter == 'playlists':
            params = 'Eg-KAQwIABAAGAAgACgB'
            if not ignore_spelling:
                params += 'MABqChAEEAMQCRAFEAo%3D'
            else:
                params += 'MABCAggBagoQBBADEAkQBRAK'

        elif 'playlists' in filter:
            param1 = 'EgeKAQQoA'
            if filter == 'featured_playlists':
                param2 = 'Dg'
            else:  # community_playlists
                param2 = 'EA'

            if not ignore_spelling:
                param3 = 'BagwQDhAKEAMQBBAJEAU%3D'
            else:
                param3 = 'BQgIIAWoMEA4QChADEAQQCRAF'

        else:
            param1 = filtered_param1
            param2 = _get_param2(filter)
            if not ignore_spelling:
                param3 = 'AWoMEA4QChADEAQQCRAF'
            else:
                param3 = 'AUICCAFqDBAOEAoQAxAEEAkQBQ%3D%3D'

    if not scope and not filter and ignore_spelling:
        params = 'EhGKAQ4IARABGAEgASgAOAFAAUICCAE%3D'

    return params if params else param1 + param2 + param3


def _get_param2(filter):
    filter_params = {
        'songs': 'II',
        'videos': 'IQ',
        'albums': 'IY',
        'artists': 'Ig',
        'playlists': 'Io',
        'profiles': 'JY'
    }
    return filter_params[filter]


def parse_search_suggestions(results, detailed_runs):
    if not results.get('contents', [{}])[0].get('searchSuggestionsSectionRenderer', {}).get(
            'contents', []):
        return []

    raw_suggestions = results['contents'][0]['searchSuggestionsSectionRenderer']['contents']
    suggestions = []

    for raw_suggestion in raw_suggestions:
        suggestion_content = raw_suggestion['searchSuggestionRenderer']

        text = suggestion_content['navigationEndpoint']['searchEndpoint']['query']
        runs = suggestion_content['suggestion']['runs']

        if detailed_runs:
            suggestions.append({'text': text, 'runs': runs})
        else:
            suggestions.append(text)

    return suggestions



from .songs import *
from ._utils import *


def parse_mixed_content(rows):
    items = []
    for row in rows:
        if DESCRIPTION_SHELF[0] in row:
            results = nav(row, DESCRIPTION_SHELF)
            title = nav(results, ['header'] + RUN_TEXT)
            contents = nav(results, DESCRIPTION)
        else:
            results = next(iter(row.values()))
            if 'contents' not in results:
                continue
            title = nav(results, CAROUSEL_TITLE + ['text'])
            contents = []
            for result in results['contents']:
                data = nav(result, [MTRIR], True)
                content = None
                if data:
                    page_type = nav(data, TITLE + NAVIGATION_BROWSE + PAGE_TYPE, True)
                    if page_type is None:  # song or watch_playlist
                        if nav(data, NAVIGATION_WATCH_PLAYLIST_ID, True) is not None:
                            content = parse_watch_playlist(data)
                        else:
                            content = parse_song(data)
                    elif page_type == "MUSIC_PAGE_TYPE_ALBUM":
                        content = parse_album(data)
                    elif page_type == "MUSIC_PAGE_TYPE_ARTIST":
                        content = parse_related_artist(data)
                    elif page_type == "MUSIC_PAGE_TYPE_PLAYLIST":
                        content = parse_playlist(data)
                else:
                    data = nav(result, [MRLIR])
                    content = parse_song_flat(data)

                contents.append(content)

        items.append({'title': title, 'contents': contents})
    return items


def parse_content_list(results, parse_func, key=MTRIR):
    contents = []
    for result in results:
        contents.append(parse_func(result[key]))

    return contents


def parse_album(result):
    return {
        'title': nav(result, TITLE_TEXT),
        'year': nav(result, SUBTITLE2, True),
        'browseId': nav(result, TITLE + NAVIGATION_BROWSE_ID),
        'thumbnails': nav(result, THUMBNAIL_RENDERER),
        'isExplicit': nav(result, SUBTITLE_BADGE_LABEL, True) is not None
    }


def parse_single(result):
    return {
        'title': nav(result, TITLE_TEXT),
        'year': nav(result, SUBTITLE, True),
        'browseId': nav(result, TITLE + NAVIGATION_BROWSE_ID),
        'thumbnails': nav(result, THUMBNAIL_RENDERER)
    }


def parse_song(result):
    song = {
        'title': nav(result, TITLE_TEXT),
        'videoId': nav(result, NAVIGATION_VIDEO_ID),
        'playlistId': nav(result, NAVIGATION_PLAYLIST_ID, True),
        'thumbnails': nav(result, THUMBNAIL_RENDERER)
    }
    song.update(parse_song_runs(nav(result, SUBTITLE_RUNS)))
    return song


def parse_song_flat(data):
    columns = [get_flex_column_item(data, i) for i in range(0, len(data['flexColumns']))]
    song = {
        'title': nav(columns[0], TEXT_RUN_TEXT),
        'videoId': nav(columns[0], TEXT_RUN + NAVIGATION_VIDEO_ID, True),
        'artists': parse_song_artists(data, 1),
        'thumbnails': nav(data, THUMBNAILS),
        'isExplicit': nav(data, BADGE_LABEL, True) is not None
    }
    if len(columns) > 2 and columns[2] is not None and 'navigationEndpoint' in nav(
            columns[2], TEXT_RUN):
        song['album'] = {
            'name': nav(columns[2], TEXT_RUN_TEXT),
            'id': nav(columns[2], TEXT_RUN + NAVIGATION_BROWSE_ID)
        }
    else:
        song['views'] = nav(columns[1], ['text', 'runs', -1, 'text']).split(' ')[0]

    return song


def parse_video(result):
    runs = nav(result, SUBTITLE_RUNS)
    artists_len = get_dot_separator_index(runs)
    videoId = nav(result, NAVIGATION_VIDEO_ID, True)
    if not videoId:
        videoId = nav(result, NAVIGATION_VIDEO_ID_2)
    return {
        'title': nav(result, TITLE_TEXT),
        'videoId': videoId,
        'artists': parse_song_artists_runs(runs[:artists_len]),
        'playlistId': nav(result, NAVIGATION_PLAYLIST_ID, True),
        'thumbnails': nav(result, THUMBNAIL_RENDERER, True),
        'views': runs[-1]['text'].split(' ')[0]
    }


def parse_playlist(data):
    playlist = {
        'title': nav(data, TITLE_TEXT),
        'playlistId': nav(data, TITLE + NAVIGATION_BROWSE_ID)[2:],
        'thumbnails': nav(data, THUMBNAIL_RENDERER)
    }
    subtitle = data['subtitle']
    if 'runs' in subtitle:
        playlist['description'] = "".join([run['text'] for run in subtitle['runs']])
        if len(subtitle['runs']) == 3 and re.search(r'\d+ ', nav(data, SUBTITLE2)):
            playlist['count'] = nav(data, SUBTITLE2).split(' ')[0]
            playlist['author'] = parse_song_artists_runs(subtitle['runs'][:1])

    return playlist


def parse_related_artist(data):
    subscribers = nav(data, SUBTITLE, True)
    if subscribers:
        subscribers = subscribers.split(' ')[0]
    return {
        'title': nav(data, TITLE_TEXT),
        'browseId': nav(data, TITLE + NAVIGATION_BROWSE_ID),
        'subscribers': subscribers,
        'thumbnails': nav(data, THUMBNAIL_RENDERER),
    }


def parse_watch_playlist(data):
    return {
        'title': nav(data, TITLE_TEXT),
        'playlistId': nav(data, NAVIGATION_WATCH_PLAYLIST_ID),
        'thumbnails': nav(data, THUMBNAIL_RENDERER),
    }



import json
import time
import webbrowser
from typing import Dict, Optional

import requests
from requests.structures import CaseInsensitiveDict

from ytmusicapi.constants import (OAUTH_CLIENT_ID, OAUTH_CLIENT_SECRET, OAUTH_CODE_URL,
                                  OAUTH_SCOPE, OAUTH_TOKEN_URL, OAUTH_USER_AGENT)
from ytmusicapi.helpers import initialize_headers


def is_oauth(headers: CaseInsensitiveDict) -> bool:
    oauth_structure = {
        "access_token",
        "expires_at",
        "expires_in",
        "token_type",
        "refresh_token",
    }
    return all(key in headers for key in oauth_structure)


def is_custom_oauth(headers: CaseInsensitiveDict) -> bool:
    """Checks whether the headers contain a Bearer token, indicating a custom OAuth implementation."""
    return "authorization" in headers and headers["authorization"].startswith("Bearer ")


class YTMusicOAuth:
    """OAuth implementation for YouTube Music based on YouTube TV"""

    def __init__(self, session: requests.Session, proxies: Dict = None):
        self._session = session
        if proxies:
            self._session.proxies.update(proxies)

    def _send_request(self, url, data) -> requests.Response:
        data.update({"client_id": OAUTH_CLIENT_ID})
        headers = {"User-Agent": OAUTH_USER_AGENT}
        return self._session.post(url, data, headers=headers)

    def get_code(self) -> Dict:
        code_response = self._send_request(OAUTH_CODE_URL, data={"scope": OAUTH_SCOPE})
        response_json = code_response.json()
        return response_json

    @staticmethod
    def _parse_token(response) -> Dict:
        token = response.json()
        token["expires_at"] = int(time.time()) + int(token["expires_in"])
        return token

    def get_token_from_code(self, device_code: str) -> Dict:
        response = self._send_request(
            OAUTH_TOKEN_URL,
            data={
                "client_secret": OAUTH_CLIENT_SECRET,
                "grant_type": "http://oauth.net/grant_type/device/1.0",
                "code": device_code,
            },
        )
        return self._parse_token(response)

    def refresh_token(self, refresh_token: str) -> Dict:
        response = self._send_request(
            OAUTH_TOKEN_URL,
            data={
                "client_secret": OAUTH_CLIENT_SECRET,
                "grant_type": "refresh_token",
                "refresh_token": refresh_token,
            },
        )
        return self._parse_token(response)

    @staticmethod
    def dump_token(token: Dict, filepath: Optional[str]):
        if not filepath or len(filepath) > 255:
            return
        with open(filepath, encoding="utf8", mode="w") as file:
            json.dump(token, file, indent=True)

    def setup(self, filepath: Optional[str] = None, open_browser: bool = False) -> Dict:
        code = self.get_code()
        url = f"{code['verification_url']}?user_code={code['user_code']}"
        if open_browser:
            webbrowser.open(url)
        input(f"Go to {url}, finish the login flow and press Enter when done, Ctrl-C to abort")
        token = self.get_token_from_code(code["device_code"])
        self.dump_token(token, filepath)

        return token

    def load_headers(self, token: Dict, filepath: Optional[str] = None):
        headers = initialize_headers()
        if time.time() > token["expires_at"] - 3600:
            token.update(self.refresh_token(token["refresh_token"]))
            self.dump_token(token, filepath)
        headers["Authorization"] = f"{token['token_type']} {token['access_token']}"
        headers["Content-Type"] = "application/json"
        headers["X-Goog-Request-Time"] = str(int(time.time()))
        return headers



from ytmusicapi.navigation import nav


def get_continuations(results,
                      continuation_type,
                      limit,
                      request_func,
                      parse_func,
                      ctoken_path="",
                      reloadable=False):
    items = []
    while 'continuations' in results and (limit is None or len(items) < limit):
        additionalParams = get_reloadable_continuation_params(results) if reloadable \
            else get_continuation_params(results, ctoken_path)
        response = request_func(additionalParams)
        if 'continuationContents' in response:
            results = response['continuationContents'][continuation_type]
        else:
            break
        contents = get_continuation_contents(results, parse_func)
        if len(contents) == 0:
            break
        items.extend(contents)

    return items


def get_validated_continuations(results,
                                continuation_type,
                                limit,
                                per_page,
                                request_func,
                                parse_func,
                                ctoken_path=""):
    items = []
    while 'continuations' in results and len(items) < limit:
        additionalParams = get_continuation_params(results, ctoken_path)
        wrapped_parse_func = lambda raw_response: get_parsed_continuation_items(
            raw_response, parse_func, continuation_type)
        validate_func = lambda parsed: validate_response(parsed, per_page, limit, len(items))

        response = resend_request_until_parsed_response_is_valid(request_func, additionalParams,
                                                                 wrapped_parse_func, validate_func,
                                                                 3)
        results = response['results']
        items.extend(response['parsed'])

    return items


def get_parsed_continuation_items(response, parse_func, continuation_type):
    results = response['continuationContents'][continuation_type]
    return {'results': results, 'parsed': get_continuation_contents(results, parse_func)}


def get_continuation_params(results, ctoken_path=''):
    ctoken = nav(results,
                 ['continuations', 0, 'next' + ctoken_path + 'ContinuationData', 'continuation'])
    return get_continuation_string(ctoken)


def get_reloadable_continuation_params(results):
    ctoken = nav(results, ['continuations', 0, 'reloadContinuationData', 'continuation'])
    return get_continuation_string(ctoken)


def get_continuation_string(ctoken):
    return "&ctoken=" + ctoken + "&continuation=" + ctoken


def get_continuation_contents(continuation, parse_func):
    for term in ['contents', 'items']:
        if term in continuation:
            return parse_func(continuation[term])

    return []


def resend_request_until_parsed_response_is_valid(request_func, request_additional_params,
                                                  parse_func, validate_func, max_retries):
    response = request_func(request_additional_params)
    parsed_object = parse_func(response)
    retry_counter = 0
    while not validate_func(parsed_object) and retry_counter < max_retries:
        response = request_func(request_additional_params)
        attempt = parse_func(response)
        if len(attempt['parsed']) > len(parsed_object['parsed']):
            parsed_object = attempt
        retry_counter += 1

    return parsed_object


def validate_response(response, per_page, limit, current_count):
    remaining_items_count = limit - current_count
    expected_items_count = min(per_page, remaining_items_count)

    # response is invalid, if it has less items then minimal expected count
    return len(response['parsed']) >= expected_items_count



from .playlists import parse_playlist_items
from .songs import parse_song_runs
from ._utils import *
from ytmusicapi.continuations import get_continuations


def parse_artists(results, uploaded=False):
    artists = []
    for result in results:
        data = result[MRLIR]
        artist = {}
        artist['browseId'] = nav(data, NAVIGATION_BROWSE_ID)
        artist['artist'] = get_item_text(data, 0)
        parse_menu_playlists(data, artist)
        if uploaded:
            artist['songs'] = get_item_text(data, 1).split(' ')[0]
        else:
            subtitle = get_item_text(data, 1)
            if subtitle:
                artist['subscribers'] = subtitle.split(' ')[0]
        artist['thumbnails'] = nav(data, THUMBNAILS, True)
        artists.append(artist)

    return artists


def parse_library_albums(response, request_func, limit):
    results = get_library_contents(response, GRID)
    if results is None:
        return []
    albums = parse_albums(results['items'])

    if 'continuations' in results:
        parse_func = lambda contents: parse_albums(contents)
        remaining_limit = None if limit is None else (limit - len(albums))
        albums.extend(
            get_continuations(results, 'gridContinuation', remaining_limit, request_func,
                              parse_func))

    return albums


def parse_albums(results):
    albums = []
    for result in results:
        data = result[MTRIR]
        album = {}
        album['browseId'] = nav(data, TITLE + NAVIGATION_BROWSE_ID)
        album['playlistId'] = nav(data, MENU_PLAYLIST_ID, none_if_absent=True)
        album['title'] = nav(data, TITLE_TEXT)
        album['thumbnails'] = nav(data, THUMBNAIL_RENDERER)

        if 'runs' in data['subtitle']:
            album['type'] = nav(data, SUBTITLE)
            album.update(parse_song_runs(data['subtitle']['runs'][2:]))

        albums.append(album)

    return albums


def parse_library_artists(response, request_func, limit):
    results = get_library_contents(response, MUSIC_SHELF)
    if results is None:
        return []
    artists = parse_artists(results['contents'])

    if 'continuations' in results:
        parse_func = lambda contents: parse_artists(contents)
        remaining_limit = None if limit is None else (limit - len(artists))
        artists.extend(
            get_continuations(results, 'musicShelfContinuation', remaining_limit, request_func,
                              parse_func))

    return artists


def parse_library_songs(response):
    results = get_library_contents(response, MUSIC_SHELF)
    return {
        'results': results,
        'parsed': parse_playlist_items(results['contents'][1:]) if results else results
    }


def get_library_contents(response, renderer):
    """
    Find library contents. This function is a bit messy now
    as it is supporting two different response types. Can be
    cleaned up once all users are migrated to the new responses.
    :param response: ytmusicapi response
    :param renderer: GRID or MUSIC_SHELF
    :return: library contents or None
    """
    section = nav(response, SINGLE_COLUMN_TAB + SECTION_LIST, True)
    contents = None
    if section is None:  # empty library
        contents = nav(response, SINGLE_COLUMN + TAB_1_CONTENT + SECTION_LIST_ITEM + renderer,
                       True)
    else:
        results = find_object_by_key(section, 'itemSectionRenderer')
        if results is None:
            contents = nav(response, SINGLE_COLUMN_TAB + SECTION_LIST_ITEM + renderer, True)
        else:
            contents = nav(results, ITEM_SECTION + renderer, True)
    return contents



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys
sys.path.insert(0, os.path.abspath('.'))
sys.path.insert(0, '../..')
from ytmusicapi import __version__  # noqa: E402

on_rtd = os.environ.get('READTHEDOCS', None) == 'True'

# -- Project information -----------------------------------------------------

project = 'ytmusicapi'
copyright = '2022, sigma67'
author = 'sigma67'

# The full version, including alpha/beta/rc tags
version = __version__
# The full version, including alpha/beta/rc tags.
release = __version__

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = ["sphinx.ext.autodoc"]

# The suffix of source filenames.
source_suffix = '.rst'

# The master toctree document.
master_doc = 'index'

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = []

html_theme = "sphinx_rtd_theme"


from ._utils import *
from .songs import *


def parse_watch_playlist(results):
    tracks = []
    PPVWR = 'playlistPanelVideoWrapperRenderer'
    PPVR = 'playlistPanelVideoRenderer'
    for result in results:
        counterpart = None
        if PPVWR in result:
            counterpart = result[PPVWR]['counterpart'][0]['counterpartRenderer'][PPVR]
            result = result[PPVWR]['primaryRenderer']
        if PPVR not in result:
            continue
        data = result[PPVR]
        if 'unplayableText' in data:
            continue

        track = parse_watch_track(data)
        if counterpart:
            track['counterpart'] = parse_watch_track(counterpart)
        tracks.append(track)

    return tracks


def parse_watch_track(data):
    feedback_tokens = like_status = None
    for item in nav(data, MENU_ITEMS):
        if TOGGLE_MENU in item:
            service = item[TOGGLE_MENU]['defaultServiceEndpoint']
            if 'feedbackEndpoint' in service:
                feedback_tokens = parse_song_menu_tokens(item)
            if 'likeEndpoint' in service:
                like_status = parse_like_status(service)

    song_info = parse_song_runs(data['longBylineText']['runs'])

    track = {
        'videoId': data['videoId'],
        'title': nav(data, TITLE_TEXT),
        'length': nav(data, ['lengthText', 'runs', 0, 'text'], True),
        'thumbnail': nav(data, THUMBNAIL),
        'feedbackTokens': feedback_tokens,
        'likeStatus': like_status,
        'videoType': nav(data, ['navigationEndpoint'] + NAVIGATION_VIDEO_TYPE, True)
    }
    track.update(song_info)
    return track


def get_tab_browse_id(watchNextRenderer, tab_id):
    if 'unselectable' not in watchNextRenderer['tabs'][tab_id]['tabRenderer']:
        return watchNextRenderer['tabs'][tab_id]['tabRenderer']['endpoint']['browseEndpoint'][
            'browseId']
    else:
        return None



import re
from datetime import date


def prepare_like_endpoint(rating):
    if rating == 'LIKE':
        return 'like/like'
    elif rating == 'DISLIKE':
        return 'like/dislike'
    elif rating == 'INDIFFERENT':
        return 'like/removelike'
    else:
        return None


def validate_order_parameter(order):
    orders = ['a_to_z', 'z_to_a', 'recently_added']
    if order and order not in orders:
        raise Exception(
            "Invalid order provided. Please use one of the following orders or leave out the parameter: "
            + ', '.join(orders))


def prepare_order_params(order):
    orders = ['a_to_z', 'z_to_a', 'recently_added']
    if order is not None:
        # determine order_params via `.contents.singleColumnBrowseResultsRenderer.tabs[0].tabRenderer.content.sectionListRenderer.contents[1].itemSectionRenderer.header.itemSectionTabbedHeaderRenderer.endItems[1].dropdownRenderer.entries[].dropdownItemRenderer.onSelectCommand.browseEndpoint.params` of `/youtubei/v1/browse` response
        order_params = ['ggMGKgQIARAA', 'ggMGKgQIARAB', 'ggMGKgQIABAB']
        return order_params[orders.index(order)]


def html_to_txt(html_text):
    tags = re.findall("<[^>]+>", html_text)
    for tag in tags:
        html_text = html_text.replace(tag, '')
    return html_text


def get_datestamp():
    return (date.today() - date.fromtimestamp(0)).days



from typing import List, Dict

from ytmusicapi.navigation import nav, CAROUSEL, CAROUSEL_TITLE, NAVIGATION_BROWSE_ID
from ytmusicapi.parsers._utils import i18n
from ytmusicapi.parsers.browsing import parse_album, parse_single, parse_video, parse_playlist, parse_related_artist, \
    parse_content_list


class Parser:

    def __init__(self, language):
        self.lang = language

    @i18n
    def get_search_result_types(self):
        return [_('artist'), _('playlist'), _('song'), _('video'), _('station'), _('profile')]

    @i18n
    def parse_artist_contents(self, results: List) -> Dict:
        categories = ['albums', 'singles', 'videos', 'playlists', 'related']
        categories_local = [_('albums'), _('singles'), _('videos'), _('playlists'), _('related')]
        categories_parser = [
            parse_album, parse_single, parse_video, parse_playlist, parse_related_artist
        ]
        artist = {}
        for i, category in enumerate(categories):
            data = [
                r['musicCarouselShelfRenderer'] for r in results
                if 'musicCarouselShelfRenderer' in r
                and nav(r, CAROUSEL + CAROUSEL_TITLE)['text'].lower() == categories_local[i]
            ]
            if len(data) > 0:
                artist[category] = {'browseId': None, 'results': []}
                if 'navigationEndpoint' in nav(data[0], CAROUSEL_TITLE):
                    artist[category]['browseId'] = nav(data[0],
                                                       CAROUSEL_TITLE + NAVIGATION_BROWSE_ID)
                    if category in ['albums', 'singles', 'playlists']:
                        artist[category]['params'] = nav(
                            data[0],
                            CAROUSEL_TITLE)['navigationEndpoint']['browseEndpoint']['params']

                artist[category]['results'] = parse_content_list(data[0]['contents'],
                                                                 categories_parser[i])

        return artist



from functools import wraps

from ytmusicapi.navigation import *


def parse_menu_playlists(data, result):
    watch_menu = find_objects_by_key(nav(data, MENU_ITEMS), 'menuNavigationItemRenderer')
    for item in [_x['menuNavigationItemRenderer'] for _x in watch_menu]:
        icon = nav(item, ['icon', 'iconType'])
        if icon == 'MUSIC_SHUFFLE':
            watch_key = 'shuffleId'
        elif icon == 'MIX':
            watch_key = 'radioId'
        else:
            continue

        watch_id = nav(item, ['navigationEndpoint', 'watchPlaylistEndpoint', 'playlistId'], True)
        if not watch_id:
            watch_id = nav(item, ['navigationEndpoint', 'watchEndpoint', 'playlistId'], True)
        if watch_id:
            result[watch_key] = watch_id


def get_item_text(item, index, run_index=0, none_if_absent=False):
    column = get_flex_column_item(item, index)
    if not column:
        return None
    if none_if_absent and len(column['text']['runs']) < run_index + 1:
        return None
    return column['text']['runs'][run_index]['text']


def get_flex_column_item(item, index):
    if len(item['flexColumns']) <= index or \
            'text' not in item['flexColumns'][index]['musicResponsiveListItemFlexColumnRenderer'] or \
            'runs' not in item['flexColumns'][index]['musicResponsiveListItemFlexColumnRenderer']['text']:
        return None

    return item['flexColumns'][index]['musicResponsiveListItemFlexColumnRenderer']


def get_fixed_column_item(item, index):
    if 'text' not in item['fixedColumns'][index]['musicResponsiveListItemFixedColumnRenderer'] or \
            'runs' not in item['fixedColumns'][index]['musicResponsiveListItemFixedColumnRenderer']['text']:
        return None

    return item['fixedColumns'][index]['musicResponsiveListItemFixedColumnRenderer']


def get_browse_id(item, index):
    if 'navigationEndpoint' not in item['text']['runs'][index]:
        return None
    else:
        return nav(item['text']['runs'][index], NAVIGATION_BROWSE_ID)


def get_dot_separator_index(runs):
    index = len(runs)
    try:
        index = runs.index({'text': ' • '})
    except ValueError:
        len(runs)
    return index


def parse_duration(duration):
    if duration is None:
        return duration
    mapped_increments = zip([1, 60, 3600], reversed(duration.split(":")))
    seconds = sum(multiplier * int(time) for multiplier, time in mapped_increments)
    return seconds


def i18n(method):
    @wraps(method)
    def _impl(self, *method_args, **method_kwargs):
        method.__globals__['_'] = self.lang.gettext
        return method(self, *method_args, **method_kwargs)

    return _impl



from typing import List, Dict, Union

from ytmusicapi.continuations import get_continuations
from ytmusicapi.parsers.playlists import validate_playlist_id
from ytmusicapi.parsers.watch import *


class WatchMixin:
    def get_watch_playlist(self,
                           videoId: str = None,
                           playlistId: str = None,
                           limit=25,
                           radio: bool = False,
                           shuffle: bool = False) -> Dict[str, Union[List[Dict]]]:
        """
        Get a watch list of tracks. This watch playlist appears when you press
        play on a track in YouTube Music.

        Please note that the `INDIFFERENT` likeStatus of tracks returned by this
        endpoint may be either `INDIFFERENT` or `DISLIKE`, due to ambiguous data
        returned by YouTube Music.

        :param videoId: videoId of the played video
        :param playlistId: playlistId of the played playlist or album
        :param limit: minimum number of watch playlist items to return
        :param radio: get a radio playlist (changes each time)
        :param shuffle: shuffle the input playlist. only works when the playlistId parameter
            is set at the same time. does not work if radio=True
        :return: List of watch playlist items. The counterpart key is optional and only
            appears if a song has a corresponding video counterpart (UI song/video
            switcher).

        Example::

            {
                "tracks": [
                    {
                      "videoId": "9mWr4c_ig54",
                      "title": "Foolish Of Me (feat. Jonathan Mendelsohn)",
                      "length": "3:07",
                      "thumbnail": [
                        {
                          "url": "https://lh3.googleusercontent.com/ulK2YaLtOW0PzcN7ufltG6e4ae3WZ9Bvg8CCwhe6LOccu1lCKxJy2r5AsYrsHeMBSLrGJCNpJqXgwczk=w60-h60-l90-rj",
                          "width": 60,
                          "height": 60
                        }...
                      ],
                      "feedbackTokens": {
                        "add": "AB9zfpIGg9XN4u2iJ...",
                        "remove": "AB9zfpJdzWLcdZtC..."
                      },
                      "likeStatus": "INDIFFERENT",
                      "videoType": "MUSIC_VIDEO_TYPE_ATV",
                      "artists": [
                        {
                          "name": "Seven Lions",
                          "id": "UCYd2yzYRx7b9FYnBSlbnknA"
                        },
                        {
                          "name": "Jason Ross",
                          "id": "UCVCD9Iwnqn2ipN9JIF6B-nA"
                        },
                        {
                          "name": "Crystal Skies",
                          "id": "UCTJZESxeZ0J_M7JXyFUVmvA"
                        }
                      ],
                      "album": {
                        "name": "Foolish Of Me",
                        "id": "MPREb_C8aRK1qmsDJ"
                      },
                      "year": "2020",
                      "counterpart": {
                        "videoId": "E0S4W34zFMA",
                        "title": "Foolish Of Me [ABGT404] (feat. Jonathan Mendelsohn)",
                        "length": "3:07",
                        "thumbnail": [...],
                        "feedbackTokens": null,
                        "likeStatus": "LIKE",
                        "artists": [
                          {
                            "name": "Jason Ross",
                            "id": null
                          },
                          {
                            "name": "Seven Lions",
                            "id": null
                          },
                          {
                            "name": "Crystal Skies",
                            "id": null
                          }
                        ],
                        "views": "6.6K"
                      }
                    },...
                ],
                "playlistId": "RDAMVM4y33h81phKU",
                "lyrics": "MPLYt_HNNclO0Ddoc-17"
            }

        """
        body = {
            'enablePersistentPlaylistPanel': True,
            'isAudioOnly': True,
            'tunerSettingValue': 'AUTOMIX_SETTING_NORMAL'
        }
        if not videoId and not playlistId:
            raise Exception("You must provide either a video id, a playlist id, or both")
        if videoId:
            body['videoId'] = videoId
            if not playlistId:
                playlistId = "RDAMVM" + videoId
            if not (radio or shuffle):
                body['watchEndpointMusicSupportedConfigs'] = {
                    'watchEndpointMusicConfig': {
                        'hasPersistentPlaylistPanel': True,
                        'musicVideoType': "MUSIC_VIDEO_TYPE_ATV",
                    }
                }
        body['playlistId'] = validate_playlist_id(playlistId)
        is_playlist = body['playlistId'].startswith('PL') or \
                      body['playlistId'].startswith('OLA')
        if shuffle and playlistId is not None:
            body['params'] = "wAEB8gECKAE%3D"
        if radio:
            body['params'] = "wAEB"
        endpoint = 'next'
        response = self._send_request(endpoint, body)
        watchNextRenderer = nav(response, [
            'contents', 'singleColumnMusicWatchNextResultsRenderer', 'tabbedRenderer',
            'watchNextTabbedResultsRenderer'
        ])

        lyrics_browse_id = get_tab_browse_id(watchNextRenderer, 1)
        related_browse_id = get_tab_browse_id(watchNextRenderer, 2)

        results = nav(watchNextRenderer,
                      TAB_CONTENT + ['musicQueueRenderer', 'content', 'playlistPanelRenderer'])
        playlist = next(
            filter(
                bool,
                map(
                    lambda x: nav(x, ['playlistPanelVideoRenderer'] + NAVIGATION_PLAYLIST_ID, True
                                  ), results['contents'])), None)
        tracks = parse_watch_playlist(results['contents'])

        if 'continuations' in results:
            request_func = lambda additionalParams: self._send_request(
                endpoint, body, additionalParams)
            parse_func = lambda contents: parse_watch_playlist(contents)
            tracks.extend(
                get_continuations(results, 'playlistPanelContinuation', limit - len(tracks),
                                  request_func, parse_func, '' if is_playlist else 'Radio'))

        return dict(tracks=tracks,
                    playlistId=playlist,
                    lyrics=lyrics_browse_id,
                    related=related_browse_id)



from typing import List
from .songs import *
from ._utils import *


def parse_playlist_items(results, menu_entries: List[List] = None):
    songs = []
    for result in results:
        if MRLIR not in result:
            continue
        data = result[MRLIR]

        videoId = setVideoId = None
        like = None
        feedback_tokens = None

        # if the item has a menu, find its setVideoId
        if 'menu' in data:
            for item in nav(data, MENU_ITEMS):
                if 'menuServiceItemRenderer' in item:
                    menu_service = nav(item, MENU_SERVICE)
                    if 'playlistEditEndpoint' in menu_service:
                        setVideoId = menu_service['playlistEditEndpoint']['actions'][0][
                            'setVideoId']
                        videoId = menu_service['playlistEditEndpoint']['actions'][0][
                            'removedVideoId']

                if TOGGLE_MENU in item:
                    feedback_tokens = parse_song_menu_tokens(item)

        # if item is not playable, the videoId was retrieved above
        if nav(data, PLAY_BUTTON, none_if_absent=True) is not None:
            if 'playNavigationEndpoint' in nav(data, PLAY_BUTTON):
                videoId = nav(data,
                              PLAY_BUTTON)['playNavigationEndpoint']['watchEndpoint']['videoId']

                if 'menu' in data:
                    like = nav(data, MENU_LIKE_STATUS, True)

        title = get_item_text(data, 0)
        if title == 'Song deleted':
            continue

        artists = parse_song_artists(data, 1)

        album = parse_song_album(data, 2)

        duration = None
        if 'fixedColumns' in data:
            if 'simpleText' in get_fixed_column_item(data, 0)['text']:
                duration = get_fixed_column_item(data, 0)['text']['simpleText']
            else:
                duration = get_fixed_column_item(data, 0)['text']['runs'][0]['text']

        thumbnails = None
        if 'thumbnail' in data:
            thumbnails = nav(data, THUMBNAILS)

        isAvailable = True
        if 'musicItemRendererDisplayPolicy' in data:
            isAvailable = data[
                'musicItemRendererDisplayPolicy'] != 'MUSIC_ITEM_RENDERER_DISPLAY_POLICY_GREY_OUT'

        isExplicit = nav(data, BADGE_LABEL, True) is not None

        videoType = nav(
            data, MENU_ITEMS + [0, 'menuNavigationItemRenderer', 'navigationEndpoint']
            + NAVIGATION_VIDEO_TYPE, True)

        song = {
            'videoId': videoId,
            'title': title,
            'artists': artists,
            'album': album,
            'likeStatus': like,
            'thumbnails': thumbnails,
            'isAvailable': isAvailable,
            'isExplicit': isExplicit,
            'videoType': videoType
        }
        if duration:
            song['duration'] = duration
            song['duration_seconds'] = parse_duration(duration)
        if setVideoId:
            song['setVideoId'] = setVideoId
        if feedback_tokens:
            song['feedbackTokens'] = feedback_tokens

        if menu_entries:
            for menu_entry in menu_entries:
                song[menu_entry[-1]] = nav(data, MENU_ITEMS + menu_entry)

        songs.append(song)

    return songs


def validate_playlist_id(playlistId):
    return playlistId if not playlistId.startswith("VL") else playlistId[2:]



from ytmusicapi.ytmusic import YTMusic
from ytmusicapi.setup import setup, setup_oauth
from importlib.metadata import version, PackageNotFoundError

try:
    __version__ = version("ytmusicapi")
except PackageNotFoundError:
    # package is not installed
    pass

__copyright__ = 'Copyright 2023 sigma67'
__license__ = 'MIT'
__title__ = 'ytmusicapi'
__all__ = ["YTMusic", "setup_oauth", "setup"]



import json
import os
from typing import Dict, Optional

import requests
from requests.structures import CaseInsensitiveDict

from ytmusicapi.auth.browser import is_browser
from ytmusicapi.auth.oauth import YTMusicOAuth, is_oauth, is_custom_oauth
from ytmusicapi.helpers import initialize_headers


def load_headers_file(auth: str) -> Dict:
    if os.path.isfile(auth):
        with open(auth) as json_file:
            input_json = json.load(json_file)
    else:
        input_json = json.loads(auth)

    return input_json


def prepare_headers(
    session: requests.Session,
    proxies: Optional[Dict] = None,
    input_dict: Optional[CaseInsensitiveDict] = None,
) -> Dict:
    if input_dict:

        if is_oauth(input_dict):
            oauth = YTMusicOAuth(session, proxies)
            headers = oauth.load_headers(dict(input_dict), input_dict['filepath'])

        elif is_browser(input_dict):
            headers = input_dict

        elif is_custom_oauth(input_dict):
            headers = input_dict

        else:
            raise Exception(
                "Could not detect credential type. "
                "Please ensure your oauth or browser credentials are set up correctly.")

    else:  # no authentication
        headers = initialize_headers()

    return headers
